''Module format.spin: Formatted output dirver.
''
''Copyright (C) 2006 Philip C. Pilgrim
'' 
''This program is free software; you can redistribute it and/or modify
''it under the terms of the GNU General Public License as published by
''the Free Software Foundation; either version 2 of the License, or
''(at your option) any later version.
''
''This program is distributed in the hope that it will be useful,
''but WITHOUT ANY WARRANTY; without even the implied warranty of
''MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
''GNU General Public License for more details.
''
''You should have received a copy of the GNU General Public License
''along with this program; if not, write to the Free Software
''Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
''
''CONTACT
''
''  phil@phipi.com (First contact may require addition to my whitelist.)
''
''VERSION HISTORY
''
''  0.01 alpha: released 2006.03.22 Originally released as print.spin for serial output.
''  0.02 alpha: released 2006.06.26 Added support for callbacks, in place of dedicated serial output.
''                                  Deleted routines exclusive to Parallax DEBUG window.
''
''SUMMARY
''
''  This object provides C-like formatted output to a character-based output driver.
''  It works with callbacks. To start, the caller must call the "use" routine with a valid
''  callback token. (See demo program for an example of such a token.) The routine pointed to
''  by the token must have a single parameter, which is a character to be output. "use' can
''  be called at any time to change the current output device.
''
''  IMPORTANT NOTE: This object, due to its calling depth, requires a rather large stack. Any cog that calls
''  these methods should be given a stack of 96 longs or more, depending on the other requirements of the cog.
''
OBJ

  callback :    "callback"

VAR

  long  current_callback          'Token for the current callback routine.

PUB use(cb_obj)

'' Establish the current callback routine. "cb_obj" must be a token created by callback.new.

  current_callback := cb_obj
  

''______________________________________________________________________________________________________
''
''                              ******** FORMATTED OUTPUT ROUTINES ********
''
'' These routines use a format string, similar to C's (and Perl's) printf, to format the expressions
'' being printed. There is a separate function for any given number of expressions from 0 to 6. More can
'' easily be added as necessary, using the same methods used below.
'' However, DOING SO WILL INCREASE THE STACK REQUIREMENTS.
''
'' The basic usage is this:
''
''   print.start(1, 0, -9600)   'Starts uart on output pin 0, 9600 baud (inverted).
''
''   a := b := c := 123
''   s := string("This is a string.")
''   
''   print.f3(string("\cHere's some data: %2.2X (hex) %5.1d (decimal) %16.16b (binary)\r"), a, b, c)
''   print.f2(string("Here's some more: %10.5c (character) %s (string)"), "A", s)
''
'' After clearing the debug screen, this will print:
''
''   Here's some data: 7B (hex)   123 (decimal) 0000000001111011 (binary)
''   Here's some more: AAAAA      (character) This is a string. (string)
''
'' The format string may contain both escaped characters (e.g. \r = carriage return) and format codes
'' (e.g. %2.2X = print an uppercase hex number in a field at least two characters wide using at least
'' two digits.)
''
'' Here are the escape codes that it recognizes:
''
''   \c =  0: Clear the debug screen (not valid with all drivers).
''   \h =  1: Home the cursor (not valid with all drivers).
''   \< =  3: Move the cursor left (not valid with all drivers).
''   \> =  4: Move the cursor right (not valid with all drivers).
''   \u =  5: Move the cursor up (not valid with all drivers).
''   \d =  6: Move the cursor down (not valid with all drivers).
''   \a =  7: Alarm (bell)
''   \k =  8: Backspace
''   \t =  9: Tab
''   \n = 10: Newline (linefeed)
''   \e = 11: Erase to the end of the line (not valid with all drivers).
''   \b = 12: Erase all below (not valid with all drivers).
''   \r = 13: Carriage return.
''   \q = 34: Double quote.
''   \xnn = $nn: Arbitraray hex character (e.g. \x41 = $41 = "A")
''   \\ = "\" Backslash character.
''
'' Here are the format codes it recognizes. w and p, below, represent decimal numbers. w is the width of
'' the field the data will be printed in. p is the minimum number of digits (characters) to print. Numerical
'' data is printed right-justified, padded with enough leading zeroes to print at least p digits.
'' Numerical output will never be truncated to fit a too-small field: the field will expand instead.
'' If w is omitted, zero is assumed, meaning expand the field as necessary.
'' If p is omitted, one is assumed, meaning print at least one digit and as many more as are necessary.
'' String and character data are printed left-justified in their fields, padded to the right with blanks
'' as necessary.
''
''  %w.pb %wb %b : Print the next argument in the list as binary data. (e.g. %16.8b %8b %b)
''  %w.pd %wd %d : Print the next argument in the list as signed decimal data. (e.g. %6.1d %6d %d)
''  %w.px %wx %x : Print the next argument in the list as lower-case hex data. (e.g. %16.8x %8x %x)
''  %w.pX %wX %X : Print the next argument in the list as upper-case hex data. (e.g. %16.8X %8X %X)
''  %w.pc %wc %c : Repeat the next argument (character) p times in a field w wide. (e.g. %10.5c %0.5c %c)
''  %w.ps %ws %s : Print the next arguement (address of zero-terminated string in a field at least w
''    wide. Truncate the string if its longer than p. Pad with blanks if shorter. (e.g. %12.4s %12s %s)
''  %rwn         : Repeat the next character in the format string (n) the number of times given
''    by the next argument in the list in a field w characters wide. p, if given, is ignored. (e.g. %10rX)
''
'' NOTE: The "*" format operand, available in C (but not Perl), is NOT used.


PUB f6(faddr, arg1, arg2, arg3, arg4, arg5, arg6) : addr

'' Formatted output with six arguments.

  addr := f3(f3(-||faddr, arg1, arg2, arg3), arg4, arg5, arg6)
  if faddr > 0
    return f0(addr)

PUB f5(faddr, arg1, arg2, arg3, arg4, arg5) : addr

'' Formatted output with five arguments.

  addr := f2(f3(-||faddr, arg1, arg2, arg3), arg4, arg5)
  if faddr > 0
    return f0(addr)

PUB f4(faddr, arg1, arg2, arg3, arg4) : addr

'' Formatted output with four arguments.

  addr := f2(f2(-||faddr, arg1, arg2), arg3, arg4)
  if faddr > 0
    return f0(addr)

PUB f3(faddr, arg1, arg2, arg3) : addr

'' Formatted output with three arguments.

  addr := f1(f2(-||faddr, arg1, arg2), arg3)
  if faddr > 0
    return f0(addr)

PUB f2(faddr, arg1, arg2) : addr

'' Formatted output with two arguments.

  addr := f1(f1(-||faddr, arg1), arg2)
  if faddr > 0
    return f0(addr)

PUB f1(faddr, arg1) | char, width, prec, decpt, flush

'' Formatted output with one argument.

  if faddr == 0
    return 0
  flush := faddr > 0
  || faddr
  repeat while (char := next_chr(@faddr)) => 0
    if char == "%"
      width := prec := decpt := 0
      repeat
        case char := next_chr(@faddr)
          -1:
            return 0
          ".":
            decpt := true
          "0" .. "9":
            if decpt
              prec := prec * 10 + char - "0"
            else
              width := width * 10 + char - "0"
          "d", "x", "X", "b", "s", "c", "r":
            case char            
              "d": decwp(arg1, width, prec)
              "x": hexwp(arg1, width, prec, 0)
              "X": hexwp(arg1, width, prec, 1)
              "b": binwp(arg1, width, prec)
              "s": strwp(arg1, width, prec)
              "c": chrwp(arg1, width, prec)
              "r":
                if (char := next_chr(@faddr)) < 0
                  return 0
                chrwp(char, width, arg1) 
            if flush
              quit
            else
              return -faddr
          other:
            putc("%")                    
            putc(char)
            quit
    else
      putc(char)
  return 0

PUB f0(str_addr) | char

'' Formatted output with no arguments, just the format string to output. Interprets escape codes.

  if str_addr := || str_addr
    repeat while (char := next_chr(@str_addr)) => 0
      putc(char)
  return 0

PRI next_chr(faddraddr) : char | acc1, acc2

'' Get the next character from the string, interpreting escape (\) sequences as necessary.

  if word[faddraddr] == 0
    return -1
  if (char := byte[word[faddraddr]++]) == "\"
    if (char := byte[word[faddraddr]++]) == "x"
      if acc1 := uc(byte[word[faddraddr]++])
        if acc2 := uc(byte[word[faddraddr]++])
          return lookdownz(acc1 : "0" .. "9", "A" .. "F") * 16 + lookdownz(acc2 : "0" .. "9", "A" .. "F")
        else
          return -1
      else
        return -1
    elseif acc1 := lookdown(char :  0, "c", "h", "<", ">", "u", "d", "a", "k", "t", "n", "e", "b", "r", "q")
      return        lookup(acc1  : -1,  0,   1,   3,   4,   5,   6,   7,   8,   9,   10,  11,  12,  13,  34)
  if char == 0
    return -1

PUB chr(char)

'' Print char as a character (same as putc)

  putc(char)

PUB chrw(char, width)

'' Print char as a character, left-justified in a field width characters wide.

  chrwp(char, width, 0)

PUB chrp(char, prec)

'' Repeat char as a character, prec (>0) times.

  chrwp(char, 0, prec)

PUB chrwp(char, width, prec)

'' Repeat char prec (>0) times, left-justified in a field at least width characters wide.
'' Pad to the right with spaces, as necessary.

  prec #>= 1
  repeat prec
    putc(char)
  if width > prec
    repeat width - prec
      putc(" ")   

PUB str(str_addr)

'' Print a zero-terminated string in as much space as it requires.

  strwp(str_addr, 0, 0)

PUB strw(str_addr, width)

'' Print a zero-terminated string, left-justified in a field at least width wide,
'' padding to the right with blanks as necessary.

  strwp(str_addr, width, 0)

PUB strp(str_addr, prec)

'' Print a zero-terminated string, truncating after prec characters if prec > 0.

  strwp(str_addr, 0, prec) 

PUB strwp(str_addr, width, prec) | len, i

'' In a field at least width wide, print a left-justified, zero-terminated string,
'' truncated after prec characters if prec > 0.

  if str_addr
    len := strsize(str_addr)
  else
    len := 0
  if prec > 0
    len <#= prec
  if len
    repeat i from 0 to len - 1
      putc(byte[str_addr + i])
  if width > len
    repeat width - len
      putc(" ")

PUB dec(value)

'' Print a signed decimal number only as big as it needs to be, but with at least one digit.

  decwp(value, 0, 0)

PUB decw(value, width)

'' Print a signed decimal number right-justified in a field at least width wide,
'' with only as many digits as necessary, but at least one.

  decwp(value, width, 0)

PUB decp(value, prec)

'' Print a signed decimal number of at least prec digits (including sign), padding with 0s,
'' using only as much space as necessary

  decwp(value, 0, prec)

PUB decwp(value, width, prec) | digits, sgn, mask

'' Print a signed decimal number of at least prec (>0) digits (including sign), padding with 0s,
'' right-justified in a field at least width characters wide.

  prec #>= 1
  if value < 0
    -value
    sgn := 1
  else
    sgn := 0
  prec -= sgn
  width -= sgn
  digits := 10
  mask := 1_000_000_000
  repeat until value => mask
    mask /= 10
    digits--
  if width > prec #> digits
    repeat width - (prec #> digits)
      putc(" ")
  if sgn
    putc("-")
  if prec > digits
    repeat prec - digits
      putc("0")
  if digits
    repeat digits
      putc(value / mask + "0")
      value //= mask
      mask /= 10 
      
PUB hex(value)

'' Print an upper-case hexadecimal number only as big as it needs to be, but with at least one digit.

  hexwp(value, 0, 0, 1)

PUB hexw(value, width)

'' Print an upper-case hexadecimal number right-justified in a field at least width wide,
'' with only as many digits as necessary, but at least one.

  hexwp(value, width, 0, 1)

PUB hexp(value, prec)

'' Print an upper-case hexadecimal number of at least prec digits, padding with 0s,
'' using only as much space as necessary

  hexwp(value, 0, prec, 1)

PUB hexwp(value, width, prec, ucase) | digits, char

'' Print a hexadecimal number, right-justified in a field at least width wide,
'' with at least prec (> 0) digits, padding with 0s as necessary.
'' Print upper-case if ucase; lower-case, otherwise.

  prec #>= 1
  digits := (>| value + 3) >> 2 
  if width > prec #> digits
    repeat width - (prec #> digits)
      putc(" ")
  if prec > digits
    repeat prec - digits
      putc("0")
  if digits
    value ->= digits << 2
    repeat digits
      char := lookupz((value <-= 4) & $0F : "0" .. "9", "a" .. "f")
      if ucase
        char := uc(char)
      putc(char)

PUB bin(value)

'' Print a binary number only as big as it needs to be, but with at least one digit.

  binwp(value, 0, 1)

PUB binw(value, width)

'' Print a binary number right-justified in a field at least width wide,
'' with only as digits as necessary.

  binwp(value, width, 1)

PUB binp(value, prec)

'' Print a binary number of at least prec digits, padding with 0s,
'' using only as much space as necessary

  binwp(value, 0, prec)

PUB binwp(value, width, prec) | digits

'' Print a binary number, right-justified in a field at least width wide,
'' with at least prec (> 0) digits, padding with 0s as necessary.

  prec #>= 1
  digits := >| value
  if width > prec #> digits
    repeat width - (prec #> digits)
      putc(" ")
  if prec > digits
    repeat prec - digits
      putc("0")
  if digits
    value ->= digits
    repeat digits
      putc((value <-= 1) & 1 + "0") 

PRI uc(char)

'' Return char as upper case.

  if char => "a" and char =< "z"
    return char - $20
  else
    return char

PUB putc(char)

'' Send a byte to the uart.

  callback.do1(current_callback, char)
 